/* Nico Keyboard for ATtiny2313 8MHz by matsubara 2007-2008.*/
/* http://nicotak.com */
/* version 7.1 */

#include <avr/io.h>
#include <avr/pgmspace.h>


#define MASK_MODE	(1<<2)
#define MASK_BEEP	(1<<3)
#define MASK_LED	(1<<4)
#define ID_DAT	(1<<5)
#define IC_CLK	(1<<6)

#define MODE_SIO 0
#define MODE_PS2 1

#define SWNUM_ALT	0	//XCb`ԍ
#define SWNUM_CTRL	1
#define SWNUM_SHIFT	2
#define SWNUM_CURSOR	60
#define SWNUM_NONE 0xff
							//LN^R[h
#define KB_BS		0x08 	//BACK SPACE
#define KB_ENTER	0x0D 	//ENTER
#define KB_ESC		0x1B 	//ESC
#define KB_RUP		0x0E 	//ROLLUP
#define KB_RDOWN	0x0F 	//ROLLDOWN
#define KB_RIGHT	0x1C 	//J[\Eֈړ 	
#define KB_LEFT		0x1D 	//J[\ֈړ 	
#define KB_UP		0x1E 	//J[\ֈړ 	
#define KB_DOWN		0x1F 	//J[\ֈړ 	

/*charactor code*/
PROGMEM unsigned char siotable[] = {
        0,  0,	 /*row=0*/
  		0,  0, /*CTRL*/
  		0,  0, /*SHIFT*/
      '1','!',
      'Q','q',
      'A','a',
      'Z','z',
        0,  0,
      '3','#',	/*row=1*/
      '2','\"',
      'E','e',
      'W','w',
      'D','d',
      'S','s',
      'C','c',
      'X','x',
      '5','%',	/*row=2*/
      '4','$',
      'T','t',
      'R','r',
      'G','g',
      'F','f',
      'B','b',
      'V','v',
      '7','\'',	/*row=3*/
      '6','&',
      'U','u',
      'Y','y',
      'J','j',
      'H','h',
      'M','m',
      'N','n',
      '9',')',	/*row=4*/
      '8','(',
      'O','o',
      'I','i',
      'L','l',
      'K','k',
      '.','>',
      ',','<',
      '-','=',	/*row=5*/
      '0','0',
      '@','`',
      'P','p',
      ':','*',
      ';','+',
      '_','_',
      '/','?',
      '^','~',	/*row=6*/
     '\\','|',
      '[','{',
    KB_BS,  0,
      ']','}',
 KB_ENTER,  0,
      ' ',  0,
        0,  0,
        0,  0,	/*row=7*/
        0,  0,
        0,  0,
        0,  0,
    KB_UP,  0,
  KB_DOWN,  0,
  KB_LEFT,  0,
 KB_RIGHT,  0
};


/*Make code*/
PROGMEM unsigned char ps2table[] = {
    0x11,	//SW00 ALT(append1)		/*row=0*/
    0x14,	//SW01 KB_CTRL
    0x12,	//SW02 KB_SHFT
    0x16,	//SW03     1
    0x15,	//SW04     Q
    0x1c,	//SW05     A
    0x1a,	//SW06     Z
    0x76,	//SW07     ESC(append2)
    0x26,	//SW08     3	/*row=1*/
    0x1e,	//SW09     2
    0x24,	//SW10     E
    0x1d,	//SW11     W
    0x23,	//SW12     D
    0x1b,	//SW13     S
    0x21,	//SW14     C
    0x22,	//SW15     X
    0x2e,	//SW16     5	/*row=2*/
    0x25,	//SW17     4
    0x2c,	//SW18     T
    0x2d,	//SW19     R
    0x34,	//SW20     G
    0x2b,	//SW21     F
    0x32,	//SW22     B
    0x2a,	//SW23     V
    0x3d,	//SW24     7	/*row=3*/
    0x36,	//SW25     6
    0x3c,	//SW26     U
    0x35,	//SW27     Y
    0x3b,	//SW28     J
    0x33,	//SW29     H
    0x3a,	//SW30     M
    0x31,	//SW31     N
    0x46,	//SW32     9	/*row=4*/
    0x3e,	//SW33     8
    0x44,	//SW34     O
    0x43,	//SW35     I
    0x4B,	//SW36     L
    0x42,	//SW37     K
    0x49,	//SW38     .
    0x41,	//SW39     ,
    0x4e,	//SW40     -	/*row=5*/
    0x45,	//SW41     0
    0x54,	//SW42     @
    0x4d,	//SW43     P
    0x52,	//SW44     :
    0x4c,	//SW45     ;
    0x51,	//SW46     _
    0x4a,	//SW47     /
    0x55,	//SW48     ^	/*row=6*/
    0x6A,	//SW49     yen |
    0x5b,	//SW50     [   {
    0x66,	//SW51    KB_BS
    0x5d,	//SW52     ]   }
    0x5a,	//SW53 KB_ENTER,
    0x29,	//SW54     space
    0x0E,	//SW55 zenkaku/hankaku (append3)
    0x0D,	//SW56     TAB(append4)	/*row=7*/
    0x83,	//SW57     F7(append5)
    0x0A,	//SW58     F8(append6)
    0x01,	//SW59     F9(append7)
    0x75,	//SW60 UP
    0x72,	//SW61 DOWN
    0x6b,	//SW62 LEFT
    0x74	//SW63 RIGHT
};

void ps2_sendbyte(unsigned char dat);
void ps2_recvbyte(void);
void timer_wait(unsigned int cnt);

//#define BAUDRATE 9600        //{[[g
#define BAUDRATE 38400        //{[[g

#ifndef F_CPU
#define   F_CPU   8000000    //CPUNbNg 8MHz
#endif

#define  PRESCALE   1        //vXP[l
#define  PRESCALECR 1        //vXP[ݒl


#define  SIO_PORT PORTD      //ISPp|[g
#define  SIO_PIN  PIND       //ISPp|[g
#define  SIO_DDR  DDRD       //ISPp|[go͐ݒ
#define  SIO_TX   1          //Mrbg

//#define  SIO_TIFR    TIFR1   //^C}1tO
#define  SIO_TIFR    TIFR   //^C}1tO
#define  SIO_MAXCNT  0x10000 //^C}ől

#define TCNT_SIO    (SIO_MAXCNT-((F_CPU/PRESCALE)/BAUDRATE))
#define TCNT_1MS    (SIO_MAXCNT-((F_CPU/PRESCALE)/1000))
#define TCNT_50US   (SIO_MAXCNT-((F_CPU/PRESCALE)/(1000000/50)))
#define TCNT_10US   (SIO_MAXCNT-((F_CPU/PRESCALE)/(1000000/10)))
#define TCNT_5US    (SIO_MAXCNT-((F_CPU/PRESCALE)/(1000000/5)))

#define TCNT_600HZ    (SIO_MAXCNT-((F_CPU/PRESCALE)/600))	//BEEP 600Hz

//void sio_bitwait(unsigned char dat);
void sio_tx(char sio_txdat);


//-----------------------------------------------------------------------
// EGCg
void timer_wait(unsigned int cnt)
{
    TCNT1 = cnt;
    SIO_TIFR |= (1 << TOV1);  // TIFR̃rbgNA

    while(!(SIO_TIFR & (1 << TOV1)));
    // TOV1rbg1ɂȂ܂
}

//void sio_bitwait(unsigned char dat)
//{
//    if(dat)
//        SIO_PORT |= (1<<SIO_TX);
//    else
//        SIO_PORT &= ~(1<<SIO_TX);
//    TCNT1 = SIO_TCNT;
//    SIO_TIFR |= (1 << TOV1);  // TIFR̃rbgNA
//
//    while(!(SIO_TIFR & (1 << TOV1)));
//    // TOV1rbg1ɂȂ܂
//}

//-----------------------------------------------------------------------
// 1oCgM
// sio_txdatFMf[^0x00-0xff

//ʐMLowo
#define SIO_MARK	SIO_PORT&=~(1<<SIO_TX)	// 1
#define SIO_SPACE	SIO_PORT|=(1<<SIO_TX)	// 0

//ʐMHIGHo
//#define SIO_MARK	SIO_PORT|=(1<<SIO_TX)	// 1
//#define SIO_SPACE	SIO_PORT&=~(1<<SIO_TX)	// 0

//void sio_tx(char sio_txdat)
//{
//    unsigned char mask = 0x01;
//
//    //X^[grbg
//    SIO_SPACE;
//    timer_wait(TCNT_SIO);
//
//    //f[^rbg0-7(ʂ瑗M)
//    while(mask != 0){    
//        if(sio_txdat & mask){
//    	    SIO_MARK;
//	    }else{
//        	SIO_SPACE;
//		}
//	    timer_wait(TCNT_SIO);
//        mask = mask<<1;
//    }
//    //Xgbvrbg
//    SIO_MARK;
//    timer_wait(TCNT_SIO);
//}


void sio_tx2(char sio_txdat)
{
    unsigned char mask = 0x01;

    //X^[grbg
    SIO_SPACE;

    TCNT1 = TCNT_SIO;
    SIO_TIFR |= (1 << TOV1);  // TIFR̃rbgNA
    while(!(SIO_TIFR & (1 << TOV1)));

    //f[^rbg0-7(ʂ瑗M)
    while(mask != 0){    
	    TCNT1 = TCNT_SIO;
		SIO_TIFR |= (1 << TOV1);  // TIFR̃rbgNA

        if(sio_txdat & mask){
    	    SIO_MARK;
	    }else{
        	SIO_SPACE;
		}
        mask = mask<<1;
    	while(!(SIO_TIFR & (1 << TOV1)));
    }
    //Xgbvrbg
    TCNT1 = TCNT_SIO;
    SIO_TIFR |= (1 << TOV1);  // TIFR̃rbgNA
    SIO_MARK;
    while(!(SIO_TIFR & (1 << TOV1)));
}


void wait_msec(int i)
{
	while(i--)
	{
		timer_wait(TCNT_1MS);
	}
}

//600Hz /2 = 300Hz
void beep(void)
{
	int i=60;	// 0.1sec

	while(i--)
	{
		PORTD ^= MASK_BEEP;
		timer_wait(TCNT_600HZ);
	}
//	PORTD &= ~MASK_BEEP;
}

//   b11 b10 b09 b08 b07 b06 b05 b04 b03 b02  b01
// --+ +-+ +-+ +-+ +-+ +-+ +-+ +-+ +-+ +-+ +-+    +---
//   | | | | | | | | | | | | | | | | | | | | |    |
//   +-+ +-+ +-+ +-+ +-+ +-+ +-+ +-+ +-+ +-+ +----+
//
// -+   +---+---+---+---+---+---+---+---+---+------+--
//  |sta|d0 |d1 |d2 |d3 |d4 |d5 |d6 |d7 |po |stop
//  +---+---+---+---+---+---+---+---+---+---+
#define  PS2_PORT PORTD      //ISPp|[g
#define  PS2_PIN  PIND      //ISPp|[g
#define  PS2_DDR  DDRD       //ISPp|[go͐ݒ
#define  PS2_DAT   0          //rbg
#define  PS2_CLK   1          //rbg

// PS2RlN^ 1bitM
void ps2_sendclk(void)
{
    timer_wait(TCNT_10US);		//10uSec
	PS2_DDR |= (1<<PS2_CLK);	//output Low
    timer_wait(TCNT_50US);		//50uSec
	PS2_DDR &= ~(1<<PS2_CLK);	//output High(nCCs[_X)
    timer_wait(TCNT_10US);		//10uSec
}

char ps2_recvbit(void)
{
	ps2_sendclk();

	return(PS2_PIN & (1<<PS2_DAT));
}

// PS2RlN^ 1bytesM
void ps2_sendbyte(unsigned char dat)
{
	char pari;
	int i,wdat;

	if((PS2_PIN & (1<<PS2_CLK))==0)
		return;	//CLKL̏ꍇA

	while((PS2_PIN & (1<<PS2_DAT))==0){		//DATL̏ꍇA(f[^Փ)
		ps2_recvbyte();
//		return;	
	}
	wdat = (int)dat << 1;	//0rbgڂ̓X^[grbg

	pari=0;
	for(i=0; i<8; i++){
		if((1<<i) & dat){
			pari++;		//rbg1̐JEg
		}
	}
	if((pari & 1)==0){
		//rbg1̐
		wdat |= (1<<9);	//rbg9(peB)=1
	}
	wdat |= (1<<10);	//rbg10(Xgbvrbg)=1

	for(i=0; i<11; i++){	//ʂ11rbgM
		if((1 << i) & wdat){
			PS2_DDR &= ~(1<<PS2_DAT);	//1(HighCs[_X)
		}else{
			PS2_DDR |= (1<<PS2_DAT);	//0
		}
		ps2_sendclk();	//send 1bit clock
		if((PS2_PIN & (1<<PS2_CLK))==0)
			break;	//CLKL̏ꍇA
	}
	PS2_DDR &= ~(1<<PS2_DAT);	//1(HighCs[_X)
	wait_msec(2);	// 2msec wait
					// *@ɂă^C~OႤ\
}

// PS2RlN^ 1bytesM
void ps2_recvbyte(void)
{
	unsigned char mask=0x01;
	unsigned char bytedata=0x0;

	if((PS2_PIN & (1<<PS2_CLK))==0)
		return;		//CLK=Low̏ꍇA

	if((PS2_PIN & (1<<PS2_DAT))!=0)
		return;		//DATHigh̏ꍇA(f[^Ȃ)

	PORTD ^= MASK_LED;	//LED DEBUG

	ps2_recvbit();	//startbit

	while(mask != 0)
	{	//ʂ8rbgM
		if(ps2_recvbit())
			bytedata |= mask;

		mask <<= 1;
	}
	ps2_recvbit();	//paritybit

	if(ps2_recvbit()){	//Stop bitHigh̏ꍇAM
		PS2_DDR |= (1<<PS2_DAT);	//DAT = Low
		ps2_sendclk();	//send 1bit clock
		PS2_DDR &= ~(1<<PS2_DAT);	//1(HighCs[_X)
	}else{
		return;	//Stop bitLow̏ꍇAt[~OG[
	}
	wait_msec(2);	//2msec
					// *@ɂă^C~OႤ\

	//Mf[^ǂƕԐM
	if(bytedata==0xff){		//ResetR}h
		ps2_sendbyte(0xFA);		//Ack
		//(300-500ms)
		wait_msec(600);	//600msec
		ps2_sendbyte(0xAA);		//BAT
	}else if(bytedata==0xf2){	//IDǂݏoR}h
		ps2_sendbyte(0xFA);		//Ack
		ps2_sendbyte(0xab);		//L[{[hID
		ps2_sendbyte(0x83);		//L[{[hID
//	}else if(bytedata==0xED){	//[hZbgR}h
	}else if(bytedata==0xEE){	//GR[R}h
		ps2_sendbyte(0xEE);		//
	}else{	//Cl[uR}h
/*	}else if(bytedata==0xF4){*/	//Cl[uR}h
		ps2_sendbyte(0xfA);		//Ack
	}
	PORTD ^= MASK_LED;	//LED
}


#define FLAG_MAKECODE 0
#define FLAG_BREAKCODE 1

void ps2_sendcode(char swnum,char flag)
{
	PGM_P p = (PGM_P)ps2table;

	if(swnum >= SWNUM_CURSOR)
		ps2_sendbyte(0xE0);		/*code(J[\L[)*/

	if(flag == FLAG_BREAKCODE)
		ps2_sendbyte(0xF0);		/*break code*/

	ps2_sendbyte(pgm_read_byte(p+swnum));		/*code*/
}

int main(void)
{
	#define REPEATMAX 1200	//Key Repeat count
	#define REPEATCYCLE 150	//Key Repeat cycle

	char colnum;
	char rownum;
	unsigned char swnum;
	unsigned char swbak=SWNUM_NONE;
	unsigned char sw2num;
	unsigned char sw2bak=SWNUM_NONE;
	char mode;
	int repeatcnt=0;
	PGM_P p;

	PORTD |= MASK_MODE;	//Pullup
	DDRD &= ~MASK_MODE;
	PORTB = 0xff;		//Pullup
	DDRD |= (MASK_LED | ID_DAT | IC_CLK | MASK_BEEP);
	PORTD |= (MASK_LED | ID_DAT);	//LED ON

	if(PIND & MASK_MODE){	//Mode sw=OFF
		// VA|[g
		SIO_DDR |= (1<<SIO_TX);
	    SIO_MARK;
		mode=MODE_SIO;
	}else{				//Mode sw=ON
		PS2_PORT &= ~((1<<PS2_DAT)|(1<<PS2_CLK));	//vAbvȂ/Lowo
		PS2_DDR &= ~((1<<PS2_DAT)|(1<<PS2_CLK));	//1(HighCs[_X)
		mode=MODE_PS2;
	}

    TCCR1A= 0;                // ^C}1 [h 
    TCCR1B= PRESCALECR;       // ^C}1 vXP[ݒ

	//WbNIC
	for(rownum=0; rownum<8; rownum++){
		PORTD |=  IC_CLK;	//(74HC164)CLK
		PORTD &= ~IC_CLK;	//(74HC164)CLK
	}
	wait_msec(500);	//500msec
	PORTD &= ~MASK_LED;	//LED off

	while(1)
	{
		swnum=SWNUM_NONE;
		sw2num=SWNUM_NONE;
		for(rownum=0; rownum<8; rownum++)
		{
			if(mode==MODE_PS2)
				ps2_recvbyte();

			//1s`8ŝ1Low
			if(rownum==0){
				PORTD &= ~ID_DAT;	//(74HC164)DAT
			}else{
				PORTD |= ID_DAT;	//(74HC164)DAT
			}
		    timer_wait(TCNT_50US);		//50uSec
			PORTD |=  IC_CLK;	//(74HC164)CLK
			PORTD &= ~IC_CLK;	//(74HC164)CLK
    		timer_wait(TCNT_50US);		//50uSec

			for(colnum=0; colnum<8; colnum++){
				if(((0x80 >> colnum)& PINB)==0){	//LowoꍇAXCb`ꂽ
					PORTD |= MASK_LED;	//LED
					swnum = colnum+(rownum *8);

					if(swnum <= SWNUM_SHIFT){	//Alt/Shift/Ctrl͓pϐɊi[
						sw2num = swnum;
						swnum = SWNUM_NONE;
					}
				}
			}
			PORTD &= ~MASK_LED;	//LED
		}

		if((mode==MODE_PS2)&&(sw2num != sw2bak)){	//Alt/Shift/CTRLON܂OFFɕω
			if(sw2num != SWNUM_NONE){
				ps2_sendcode(sw2num,FLAG_MAKECODE);		// Make code
			}else{
				ps2_sendcode(sw2bak,FLAG_BREAKCODE);	// Break Code
			}
			sw2bak = sw2num;
		}

		if(swnum == swbak ){	// Key Repeat
			if(swnum != SWNUM_NONE){	// ON --> ON
				repeatcnt++;
				if(repeatcnt > REPEATMAX){
					repeatcnt = (REPEATMAX-REPEATCYCLE);
					swbak = SWNUM_NONE;
				}
			}
		}
		if(swnum != swbak ){	//XCb`ON܂OFFɕω
			if(swnum != SWNUM_NONE){	//OFF --> ON
				if(mode==MODE_SIO){	//SIO(RS232C) mode
					p = (PGM_P)siotable +(swnum<<1);
					if(sw2num == SWNUM_SHIFT){
						p++;
					}
					sio_tx2(pgm_read_byte(p));
				}else{				// PS/2 mode
					ps2_sendcode(swnum,FLAG_MAKECODE);	//make code
				}
				beep();
			}else{	//ON --> OFF
				if(mode==MODE_PS2){	// PS/2 mode
					ps2_sendcode(swbak,FLAG_BREAKCODE);	//Break Code
				}
				repeatcnt=0;
			}
			swbak = swnum;
		}
	}
	return(0);
}
